import logging
from abc import ABC, abstractclassmethod

logger = logging.getLogger(__name__)


class LambdaEvent(ABC):
    """
    Abstract data model for handling Lambda events.

    Subclasses must implement the abstract class method ``_from_event``,
    returning an instance of the data model parsed from the loaded event and
    context. The public method ``from_event`` should then be used to parse
    events.

    Using the public method allows ``raw_event`` and ``raw_context`` to be
    accessed if needed, though doing so will emit a warning.
    """

    @property
    def raw_event(self):
        msg = (
            "Using the raw event is discouraged! "
            "Consider expanding the event's data model."
        )
        logger.warning(msg)
        return self._raw_event

    @property
    def raw_context(self):
        msg = (
            "Using the raw context is discouraged! "
            "Consider expanding the event's data model."
        )
        logger.warning(msg)
        return self._raw_context

    @classmethod
    def from_event(cls, event, context):
        """
        Return an instance of the event data model from the raw event and
        context objects.

        The event and context are stored on ``raw_event`` and ``raw_context``
        if needed.

        Parameters
        ----------
        event : dict(str, Any)
            Lambda event JSON
        context : dict(str, Any)
            Lambda context JSON
        """
        result = cls._from_event(event, context)
        result._raw_event = event
        result._raw_context = context

        return result

    @abstractclassmethod
    def _from_event(cls, event, context):
        pass


def lambda_handler(event_type):
    """
    Decorator factory for wrapping a lambda handler in a boilerplate event
    logger and parser.

    Parameters
    ----------
    event_type : LambdaEvent
        The type of event the handler is handling. The handler should accept
        this type of object.

    Returns
    -------
    handler : callable
        Decorator for the Lambda handler, accepting a LambdaEvent. It logs
        the raw incoming event and the result.
    """

    def handler(func):
        def wrapped(event, context):
            logger.info(f"Event: {event!r}")
            parsed = event_type.from_event(event, context)
            result = func(parsed)
            logger.info(f"Result: {result!r}")
            return result
        return wrapped
    return handler



import os
import unittest
from unittest import mock

import boto3

from awstin.config import aws_config
from awstin.constants import TEST_SNS_ENDPOINT
from awstin.sns.api import SNSTopic


class TestSNSTopic(unittest.TestCase):
    def setUp(self):
        config = aws_config(endpoint=os.environ.get(TEST_SNS_ENDPOINT))
        self.sns_client = boto3.client("sns", **config)

        self.addCleanup(self.cleanup_topics)

    def cleanup_topics(self):
        topics = self.sns_client.list_topics()["Topics"]
        arns = [topic["TopicArn"] for topic in topics]
        for arn in arns:
            self.sns_client.delete_topic(TopicArn=arn)

    def test_instantiate_creates_topic(self):
        SNSTopic("name_of_topic")
        topics = self.sns_client.list_topics()["Topics"]
        self.assertEqual(len(topics), 1)
        topic, = topics
        self.assertTrue(topic["TopicArn"].endswith("name_of_topic"))

    def test_post_to_sns_topic(self):
        # Mock out SNS publish - may be possible to integration test later
        # with subscriptions to lambdas.
        topic = SNSTopic("some_topic")

        mock_sns = mock.Mock()
        mock_sns.publish = mock.Mock(return_value={"MessageId": "msg_id"})

        with mock.patch.object(topic, "topic", mock_sns):
            message_id = topic.publish("a cool message")

        mock_sns.publish.assert_called_once_with(Message="a cool message")
        self.assertEqual(message_id, "msg_id")



import os

import boto3

from awstin.config import aws_config
from awstin.constants import TEST_SNS_ENDPOINT


class SNSTopic:
    """
    A client for typical use of an SNS topic
    """
    def __init__(self, topic_name):
        """
        Parameters
        ----------
        topic_name : str
            Name of the topic
        """
        config = aws_config(endpoint=os.environ.get(TEST_SNS_ENDPOINT))
        self.sns = boto3.resource("sns", **config)
        self.topic = self.sns.create_topic(Name=topic_name)

    def publish(self, message):
        """
        Publish a message to the topic

        Parameters
        ----------
        message : str
            The message to send

        Returns
        -------
        str
            The message's unique ID
        """
        return self.topic.publish(Message=message)["MessageId"]



from contextlib import ExitStack
import unittest

from awstin.dynamodb.api import DynamoDB
from awstin.dynamodb.testing import temporary_dynamodb_table


class TestDynamodb(unittest.TestCase):
    def test_dynamodb_table(self):
        with temporary_dynamodb_table("table_name", "hashkey_name") as table:
            test_item = {
                "hashkey_name": "test_value",
                "another_key": 5,
            }

            # Table can be used
            table.put_item(test_item)
            result_item = table["test_value"]
            self.assertEqual(result_item, test_item)

            # DynamoDB can access table
            dynamodb = DynamoDB()
            self.assertEqual(dynamodb["table_name"]["test_value"], test_item)

    def test_dynamodb_table_access_by_full_primary_key(self):
        with temporary_dynamodb_table("tabname", "hashname") as table:
            test_item = {"hashname": "123", "nothashname": "456"}
            table.put_item(test_item)

            self.assertEqual(table[{"hashname": "123"}], test_item)

    def test_dynamodb_table_delete_item(self):
        with temporary_dynamodb_table("tabname", "hashname") as table:
            test_item = {"hashname": "123", "nothashname": "456"}
            table.put_item(test_item)

            self.assertEqual(table["123"], test_item)

            table.delete_item("123")

            with self.assertRaises(KeyError):
                table["123"]

    def test_dynamodb_list_tables(self):
        with temporary_dynamodb_table("test_tab1", "test_key1"):
            with temporary_dynamodb_table("test_tab2", "test_key2"):
                tables = DynamoDB().tables

        self.assertEqual(tables, ["test_tab1", "test_tab2"])

    def test_dynamodb_list_tables_long(self):
        dynamodb = DynamoDB()

        table_managers = [
            temporary_dynamodb_table("abc"+str(i), "def"+str(i))
            for i in range(123)
        ]
        with ExitStack() as stack:
            for table in table_managers:
                stack.enter_context(table)

            self.assertEqual(len(dynamodb.tables), 123)



import logging
import unittest

from awstin.awslambda.api import LambdaEvent
from awstin.awslambda.api import __name__ as EVENT_NAME
from awstin.awslambda.api import lambda_handler


class CustomEvent(LambdaEvent):
    def __init__(self, request_id, memory_limit):
        self.request_id = request_id
        self.memory_limit = memory_limit

    @classmethod
    def _from_event(cls, event, context):
        request_id = event["requestContext"]["requestId"]
        memory_limit = context["memory_limit_in_mb"]
        return cls(request_id, memory_limit)


class TestLambdaEvent(unittest.TestCase):
    def test_custom_event_from_event(self):
        event = {
            "requestContext": {"requestId": "id1"},
        }
        context = {
            "memory_limit_in_mb": 9001,
        }
        custom_event = CustomEvent.from_event(event, context)

        self.assertEqual(custom_event.request_id, "id1")
        self.assertEqual(custom_event.memory_limit, 9001)

        with self.assertLogs(EVENT_NAME, logging.WARNING) as logs:
            self.assertEqual(custom_event.raw_event, event)

        self.assertEqual(len(logs.records), 1)
        record, = logs.records
        self.assertIn(
            "Using the raw event is discouraged!",
            record.getMessage()
        )

        with self.assertLogs(EVENT_NAME, logging.WARNING) as logs:
            self.assertEqual(custom_event.raw_context, context)

        self.assertEqual(len(logs.records), 1)
        record, = logs.records
        self.assertIn(
            "Using the raw context is discouraged!",
            record.getMessage()
        )


class TestEventDecorator(unittest.TestCase):
    def test_custom_event_handler(self):
        event = {
            "requestContext": {"requestId": "an ID"},
        }
        context = {
            "memory_limit_in_mb": 333,
        }

        has_been_called = False

        @lambda_handler(CustomEvent)
        def handle_custom_event(custom_event):
            self.assertIsInstance(custom_event, CustomEvent)

            self.assertEqual(custom_event.request_id, "an ID")
            self.assertEqual(custom_event.memory_limit, 333)

            nonlocal has_been_called
            has_been_called = True

            return "Result!"

        # Externally the handler takes the normal signature
        with self.assertLogs(EVENT_NAME, logging.INFO) as logs:
            result = handle_custom_event(event, context)

        self.assertTrue(has_been_called)

        self.assertEqual(result, "Result!")

        self.assertEqual(len(logs.records), 2)
        record1, record2 = logs.records

        self.assertEqual(f"Event: {event!r}", record1.getMessage())
        self.assertEqual("Result: 'Result!'", record2.getMessage())
